/*
 * Copyright (c) 2014, Oracle and/or its affiliates. All rights reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.  Oracle designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Oracle in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Oracle, 500 Oracle Parkway, Redwood Shores, CA 94065 USA
 * or visit www.oracle.com if you need additional information or have any
 * questions.
 */
package com.oracle.truffle.api.utilities;

import com.oracle.truffle.api.*;
import com.oracle.truffle.api.CompilerDirectives.*;

/**
 * Utility class to speculate on conditions to be never true or to be never false. Additionally to
 * {@link BinaryConditionProfile} this implementation of {@link ConditionProfile} also counts the
 * number of times the condition was true and false. This information is reported to the underlying
 * optimization system using {@link CompilerDirectives#injectBranchProbability(double, boolean)}.
 * Condition profiles are intended to be used as part of if conditions.
 *
 * @see ConditionProfile#createCountingProfile()
 */
public final class CountingConditionProfile extends ConditionProfile {

    @CompilationFinal private int trueCount;
    @CompilationFinal private int falseCount;

    CountingConditionProfile() {
        /* package protected constructor */
    }

    @Override
    public boolean profile(boolean value) {
        if (value) {
            if (trueCount == 0) {
                CompilerDirectives.transferToInterpreterAndInvalidate();
            }
            if (CompilerDirectives.inInterpreter()) {
                trueCount++;
            }
        } else {
            if (falseCount == 0) {
                CompilerDirectives.transferToInterpreterAndInvalidate();
            }
            if (CompilerDirectives.inInterpreter()) {
                falseCount++;
            }
        }
        return CompilerDirectives.injectBranchProbability((double) trueCount / (double) (trueCount + falseCount), value);
    }

    public int getTrueCount() {
        return trueCount;
    }

    public int getFalseCount() {
        return falseCount;
    }

    @Override
    public String toString() {
        return String.format("%s(trueCount=%s, falseCount=%s)@%x", getClass().getSimpleName(), trueCount, falseCount, hashCode());
    }
}
